from collections.abc import Sequence
from typing import Any

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pandera as pa
import shapely
from numpy.typing import NDArray
from pandera.typing import Series
from pandera.typing.geopandas import GeoSeries

from ribasim.input_base import SpatialTableModel

__all__ = ("Node",)


class NodeSchema(pa.SchemaModel):
    name: Series[str] = pa.Field(default="")
    type: Series[str] = pa.Field(default="")
    allocation_network_id: Series[pd.Int64Dtype] = pa.Field(
        default=pd.NA, nullable=True, coerce=True
    )
    geometry: GeoSeries[Any] = pa.Field(default=None, nullable=True)

    class Config:
        add_missing_columns = True


class Node(SpatialTableModel[NodeSchema]):
    """The Ribasim nodes as Point geometries."""

    @staticmethod
    def node_ids_and_types(*nodes):
        # TODO Not sure if this staticmethod belongs here
        data_types = {"node_id": int, "node_type": str}
        node_type = pd.DataFrame(
            {col: pd.Series(dtype=dtype) for col, dtype in data_types.items()}
        )

        for node in nodes:
            if not node:
                continue

            ids, types = node.node_ids_and_types()
            node_type_table = pd.DataFrame(
                data={
                    "node_id": ids,
                    "node_type": types,
                }
            )
            node_type = node_type._append(node_type_table)

        node_type = node_type.drop_duplicates(subset="node_id")
        node_type = node_type.sort_values("node_id")

        node_id = node_type.node_id.tolist()
        node_type = node_type.node_type.tolist()

        return node_id, node_type

    def geometry_from_connectivity(
        self, from_id: Sequence[int], to_id: Sequence[int]
    ) -> NDArray[Any]:
        """
        Create edge shapely geometries from connectivities.

        Parameters
        ----------
        node : Ribasim.Node
        from_id : Sequence[int]
            First node of every edge.
        to_id : Sequence[int]
            Second node of every edge.

        Returns
        -------
        edge_geometry : np.ndarray
            Array of shapely LineStrings.
        """
        geometry = self.df["geometry"]
        from_points = shapely.get_coordinates(geometry.loc[from_id])
        to_points = shapely.get_coordinates(geometry.loc[to_id])
        n = len(from_points)
        vertices = np.empty((n * 2, 2), dtype=from_points.dtype)
        vertices[0::2, :] = from_points
        vertices[1::2, :] = to_points
        indices = np.repeat(np.arange(n), 2)
        return shapely.linestrings(coords=vertices, indices=indices)

    def connectivity_from_geometry(
        self, lines: NDArray[Any]
    ) -> tuple[NDArray[Any], NDArray[Any]]:
        """
        Derive from_node_id and to_node_id for every edge in lines. LineStrings
        may be used to connect multiple nodes in a sequence, but every linestring
        vertex must also a node.

        Parameters
        ----------
        node : Node
        lines : np.ndarray
            Array of shapely linestrings.

        Returns
        -------
        from_node_id : np.ndarray of int
        to_node_id : np.ndarray of int
        """
        node_index = self.df.index
        node_xy = shapely.get_coordinates(self.df.geometry.values)
        edge_xy = shapely.get_coordinates(lines)

        xy = np.vstack([node_xy, edge_xy])
        _, inverse = np.unique(xy, return_inverse=True, axis=0)
        _, index, inverse = np.unique(
            xy, return_index=True, return_inverse=True, axis=0
        )
        uniques_index = index[inverse]

        node_node_id, edge_node_id = np.split(uniques_index, [len(node_xy)])
        if not np.isin(edge_node_id, node_node_id).all():
            raise ValueError(
                "Edge lines contain coordinates that are not in the node layer. "
                "Please ensure all edges are snapped to nodes exactly."
            )

        edge_node_id = edge_node_id.reshape((-1, 2))
        from_id = node_index[edge_node_id[:, 0]].to_numpy()
        to_id = node_index[edge_node_id[:, 1]].to_numpy()
        return from_id, to_id

    def plot(self, ax=None, zorder=None) -> Any:
        """
        Plot the nodes. Each node type is given a separate marker.

        Parameters
        ----------
        ax : Optional
            The axis on which the nodes will be plotted.

        Returns
        -------
        None
        """
        if ax is None:
            _, ax = plt.subplots()
            ax.axis("off")

        MARKERS = {
            "Basin": "o",
            "FractionalFlow": "^",
            "LevelBoundary": "o",
            "LinearResistance": "^",
            "ManningResistance": "D",
            "TabulatedRatingCurve": "D",
            "Pump": "h",
            "Outlet": "h",
            "Terminal": "s",
            "FlowBoundary": "h",
            "DiscreteControl": "*",
            "PidControl": "x",
            "User": "s",
            "": "o",
        }

        COLORS = {
            "Basin": "b",
            "FractionalFlow": "r",
            "LevelBoundary": "g",
            "LinearResistance": "g",
            "ManningResistance": "r",
            "TabulatedRatingCurve": "g",
            "Pump": "0.5",  # grayscale level
            "Outlet": "g",
            "Terminal": "m",
            "FlowBoundary": "m",
            "DiscreteControl": "k",
            "PidControl": "k",
            "User": "g",
            "": "k",
        }

        for nodetype, df in self.df.groupby("type"):
            assert isinstance(nodetype, str)
            marker = MARKERS[nodetype]
            color = COLORS[nodetype]
            ax.scatter(
                df.geometry.x,
                df.geometry.y,
                marker=marker,
                color=color,
                zorder=zorder,
                label=nodetype,
            )

        geometry = self.df["geometry"]
        for text, xy in zip(self.df.index, np.column_stack((geometry.x, geometry.y))):
            ax.annotate(text=text, xy=xy, xytext=(2.0, 2.0), textcoords="offset points")

        return ax



# generated by datamodel-codegen:
#   filename:  root.schema.json

from __future__ import annotations

from datetime import datetime

from pydantic import Field

from ribasim.input_base import BaseModel


class BasinProfile(BaseModel):
    node_id: int
    area: float
    level: float
    remarks: str = Field("", description="a hack for pandera")


class BasinState(BaseModel):
    node_id: int
    level: float
    remarks: str = Field("", description="a hack for pandera")


class BasinStatic(BaseModel):
    node_id: int
    drainage: float
    potential_evaporation: float
    infiltration: float
    precipitation: float
    urban_runoff: float
    remarks: str = Field("", description="a hack for pandera")


class BasinSubgrid(BaseModel):
    subgrid_id: int
    node_id: int
    basin_level: float
    subgrid_level: float
    remarks: str = Field("", description="a hack for pandera")


class BasinTime(BaseModel):
    node_id: int
    time: datetime
    drainage: float
    potential_evaporation: float
    infiltration: float
    precipitation: float
    urban_runoff: float
    remarks: str = Field("", description="a hack for pandera")


class DiscreteControlCondition(BaseModel):
    node_id: int
    listen_feature_id: int
    variable: str
    greater_than: float
    look_ahead: float | None = None
    remarks: str = Field("", description="a hack for pandera")


class DiscreteControlLogic(BaseModel):
    node_id: int
    truth_state: str
    control_state: str
    remarks: str = Field("", description="a hack for pandera")


class Edge(BaseModel):
    fid: int
    name: str
    from_node_id: int
    to_node_id: int
    edge_type: str
    allocation_network_id: int | None = None
    remarks: str = Field("", description="a hack for pandera")


class FlowBoundaryStatic(BaseModel):
    node_id: int
    active: bool | None = None
    flow_rate: float
    remarks: str = Field("", description="a hack for pandera")


class FlowBoundaryTime(BaseModel):
    node_id: int
    time: datetime
    flow_rate: float
    remarks: str = Field("", description="a hack for pandera")


class FractionalFlowStatic(BaseModel):
    node_id: int
    fraction: float
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class LevelBoundaryStatic(BaseModel):
    node_id: int
    active: bool | None = None
    level: float
    remarks: str = Field("", description="a hack for pandera")


class LevelBoundaryTime(BaseModel):
    node_id: int
    time: datetime
    level: float
    remarks: str = Field("", description="a hack for pandera")


class LinearResistanceStatic(BaseModel):
    node_id: int
    active: bool | None = None
    resistance: float
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class ManningResistanceStatic(BaseModel):
    node_id: int
    active: bool | None = None
    length: float
    manning_n: float
    profile_width: float
    profile_slope: float
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class Node(BaseModel):
    fid: int
    name: str
    type: str
    allocation_network_id: int | None = None
    remarks: str = Field("", description="a hack for pandera")


class OutletStatic(BaseModel):
    node_id: int
    active: bool | None = None
    flow_rate: float
    min_flow_rate: float | None = None
    max_flow_rate: float | None = None
    min_crest_level: float | None = None
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class PidControlStatic(BaseModel):
    node_id: int
    active: bool | None = None
    listen_node_id: int
    target: float
    proportional: float
    integral: float
    derivative: float
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class PidControlTime(BaseModel):
    node_id: int
    listen_node_id: int
    time: datetime
    target: float
    proportional: float
    integral: float
    derivative: float
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class PumpStatic(BaseModel):
    node_id: int
    active: bool | None = None
    flow_rate: float
    min_flow_rate: float | None = None
    max_flow_rate: float | None = None
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class TabulatedRatingCurveStatic(BaseModel):
    node_id: int
    active: bool | None = None
    level: float
    flow_rate: float
    control_state: str | None = None
    remarks: str = Field("", description="a hack for pandera")


class TabulatedRatingCurveTime(BaseModel):
    node_id: int
    time: datetime
    level: float
    flow_rate: float
    remarks: str = Field("", description="a hack for pandera")


class TerminalStatic(BaseModel):
    node_id: int
    remarks: str = Field("", description="a hack for pandera")


class UserStatic(BaseModel):
    node_id: int
    active: bool | None = None
    demand: float
    return_factor: float
    min_level: float
    priority: int
    remarks: str = Field("", description="a hack for pandera")


class UserTime(BaseModel):
    node_id: int
    time: datetime
    demand: float
    return_factor: float
    min_level: float
    priority: int
    remarks: str = Field("", description="a hack for pandera")


class Root(BaseModel):
    basinprofile: BasinProfile | None = None
    basinstate: BasinState | None = None
    basinstatic: BasinStatic | None = None
    basinsubgrid: BasinSubgrid | None = None
    basintime: BasinTime | None = None
    discretecontrolcondition: DiscreteControlCondition | None = None
    discretecontrollogic: DiscreteControlLogic | None = None
    edge: Edge | None = None
    flowboundarystatic: FlowBoundaryStatic | None = None
    flowboundarytime: FlowBoundaryTime | None = None
    fractionalflowstatic: FractionalFlowStatic | None = None
    levelboundarystatic: LevelBoundaryStatic | None = None
    levelboundarytime: LevelBoundaryTime | None = None
    linearresistancestatic: LinearResistanceStatic | None = None
    manningresistancestatic: ManningResistanceStatic | None = None
    node: Node | None = None
    outletstatic: OutletStatic | None = None
    pidcontrolstatic: PidControlStatic | None = None
    pidcontroltime: PidControlTime | None = None
    pumpstatic: PumpStatic | None = None
    tabulatedratingcurvestatic: TabulatedRatingCurveStatic | None = None
    tabulatedratingcurvetime: TabulatedRatingCurveTime | None = None
    terminalstatic: TerminalStatic | None = None
    userstatic: UserStatic | None = None
    usertime: UserTime | None = None



"""
Generate Pandera Schemas from all autogenerated Pydantic Models

These classes have Schema as a postfix, so Ribasim.models.PumpStatic
becomes Ribasim.schemas.PumpStaticSchema.
"""
import inspect
import sys

import pandera as pa
from pandera.engines.pandas_engine import PydanticModel

from ribasim import models


def gen_schema(name, cls):
    cname = f"{name}Schema"
    ctype = type(
        cname,
        (pa.DataFrameModel,),
        {
            "Config": type(
                f"{cname}.Config",
                (),
                {
                    "dtype": PydanticModel(cls),
                    "coerce": True,
                    "add_missing_columns": True,
                },
            )
        },
    )
    setattr(sys.modules[__name__], cname, ctype)


for name, cls in inspect.getmembers(models, inspect.isclass):
    gen_schema(name, cls)



from collections.abc import Sequence


def prefix_column(
    column: str, record_columns: Sequence[str], prefix: str = "meta_"
) -> str:
    """Prefix column name with `prefix` if not in record_columns."""
    if (
        len(record_columns) > 0
        and column not in record_columns
        and column != "fid"
        and not column.startswith(prefix)
    ):
        column = f"{prefix}{column}"
    return column
